"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.applyMetaManifest = applyMetaManifest;

var _fs = _interopRequireDefault(require("fs"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function applyMetaManifest(compilation, plugin) {
  const {
    moduleIdMap
  } = plugin;
  const manifestPath = './metaManifest.json';

  if (!plugin.detailedManifest || Object.keys(plugin.detailedManifest).length === 0) {
    console.warn('No detailedManifest available, skipping manifest comparison');
    return {
      extraKeys: new Set(),
      movedKeys: new Set(),
      error: 'No detailedManifest available'
    };
  } // Track keys in current build


  const currentBuildKeys = new Set();
  const currentKeyToModule = new Map();
  const currentModules = new Map();
  const currentKeySets = new Map();
  const moduleIdToName = new Map(); // Process current build modules with final IDs

  for (const [resourcePath, chunkData] of Object.entries(plugin.detailedManifest)) {
    // Get the final module ID if it exists in our mapping
    const finalModuleId = moduleIdMap.get(resourcePath) || resourcePath;
    const chunkName = chunkData.chunkName || `chunk_${chunkData.chunkId || "unknown"}`; // Store module name mapping for better logging

    moduleIdToName.set(finalModuleId, chunkName);
    const keys = Array.isArray(chunkData.keysUsed) ? chunkData.keysUsed : Array.from(chunkData.keysUsed); // Skip empty modules

    if (keys.length === 0) continue; // Sort keys for consistent comparison

    const sortedKeys = [...keys].sort();
    const keySetString = sortedKeys.join(','); // Store module data using final IDs

    currentModules.set(finalModuleId, {
      id: finalModuleId,
      resourcePath: resourcePath,
      chunkId: chunkData.chunkId,
      name: chunkName,
      keys: sortedKeys
    }); // Store key set for module fingerprinting

    currentKeySets.set(keySetString, finalModuleId); // Store individual key mappings

    keys.forEach(key => {
      currentBuildKeys.add(key);
      currentKeyToModule.set(key, finalModuleId);
    });
  }

  const currentKeyCount = currentBuildKeys.size;
  const currentModuleCount = currentModules.size;
  console.log(`Current build: ${currentKeyCount} keys in ${currentModuleCount} modules`);

  try {
    // Read from specified manifest path
    if (!_fs.default.existsSync(manifestPath)) {
      console.warn(`Meta manifest not found at ${manifestPath}`);
      return {
        extraKeys: new Set(),
        movedKeys: new Set(),
        error: `Meta manifest not found at ${manifestPath}`
      };
    }

    const fileContent = _fs.default.readFileSync(manifestPath, 'utf8');

    const remoteMetaManifest = JSON.parse(fileContent); // Handle non-existent or invalid modules

    if (!remoteMetaManifest.modules) {
      console.warn('Meta manifest does not contain modules section');
      return {
        extraKeys: new Set(),
        movedKeys: new Set(),
        error: 'Meta manifest does not contain modules section'
      };
    } // Track all keys in remote manifest modules


    const remoteModuleKeys = new Set();
    const remoteKeyToModule = new Map();
    const remoteModules = new Map();
    const remoteKeySets = new Map();
    const remoteModuleIdToName = new Map();
    const moduleEquivalenceMap = new Map(); // Process remote manifest modules

    for (const [moduleId, moduleData] of Object.entries(remoteMetaManifest.modules)) {
      if (moduleData.keys && Array.isArray(moduleData.keys)) {
        // Store module name for better logging
        const moduleName = moduleId;
        remoteModuleIdToName.set(moduleId, moduleName); // Skip empty modules

        if (moduleData.keys.length === 0) continue; // Sort keys for consistent comparison

        const sortedKeys = [...moduleData.keys].sort();
        const keySetString = sortedKeys.join(','); // Store module data

        remoteModules.set(moduleId, {
          id: moduleData.id || moduleId,
          name: moduleName,
          keys: sortedKeys,
          keyCount: moduleData.keys.length
        }); // Store key set

        remoteKeySets.set(keySetString, moduleId); // Store individual key mappings

        moduleData.keys.forEach(key => {
          remoteModuleKeys.add(key);
          remoteKeyToModule.set(key, moduleId);
        });
      }
    } // Get keys from the master chunk in the remote manifest (if it exists)


    const remoteMasterKeys = new Set();

    if (remoteMetaManifest.masterKeys && Array.isArray(remoteMetaManifest.masterKeys)) {
      remoteMetaManifest.masterKeys.forEach(key => {
        remoteMasterKeys.add(key);
      });
    } // Get all unique keys from the remote manifest (modules + master)


    const remoteAllKeys = new Set([...remoteModuleKeys, ...remoteMasterKeys]);
    const remoteKeyCount = remoteAllKeys.size;
    const remoteModuleCount = remoteModules.size;
    const remoteMasterKeyCount = remoteMasterKeys.size;
    console.log(`Remote manifest: ${remoteKeyCount} keys in ${remoteModuleCount} modules, ${remoteMasterKeyCount} master keys`); // Find intersection and differences for keys

    const inBothSets = new Set([...currentBuildKeys].filter(key => remoteAllKeys.has(key)));
    const missingKeys = new Set([...remoteAllKeys].filter(key => !currentBuildKeys.has(key)));
    const extraKeys = new Set([...currentBuildKeys].filter(key => !remoteAllKeys.has(key)));
    console.log(`Key comparison: ${inBothSets.size} shared keys, ${missingKeys.size} missing, ${extraKeys.size} new`); // Enhanced multi-level module equivalence mapping
    // 1. First pass: Match by module name

    for (const [remoteId, remoteModule] of remoteModules.entries()) {
      if (remoteModule.name) {
        // Find a current module with matching name
        for (const [currentId, currentModule] of currentModules.entries()) {
          if (currentModule.name && currentModule.name === remoteModule.name) {
            moduleEquivalenceMap.set(remoteId, currentId);
            break;
          }
        }
      }
    } // 2. Second pass: Match by resource path


    const unmatchedRemote = [...remoteModules.keys()].filter(id => !moduleEquivalenceMap.has(id));

    for (const remoteId of unmatchedRemote) {
      for (const [resourcePath, currentModuleId] of moduleIdMap.entries()) {
        if (resourcePath.includes(remoteId)) {
          moduleEquivalenceMap.set(remoteId, currentModuleId);
          break;
        }
      }
    } // 3. Third pass: Match by key set fingerprint


    const stillUnmatchedRemote = [...remoteModules.keys()].filter(id => !moduleEquivalenceMap.has(id));

    for (const remoteId of stillUnmatchedRemote) {
      const remoteModule = remoteModules.get(remoteId);
      const keySetString = remoteModule.keys.join(',');

      if (currentKeySets.has(keySetString)) {
        moduleEquivalenceMap.set(remoteId, currentKeySets.get(keySetString));
      }
    } // Find moved keys based on enhanced module equivalence map


    const trulyMovedKeys = new Set();
    const keyMovementDetails = new Map();

    for (const key of inBothSets) {
      // Skip keys that are in the master chunk
      if (remoteMasterKeys.has(key)) {
        continue;
      }

      const remoteModuleId = remoteKeyToModule.get(key);
      const currentModuleId = currentKeyToModule.get(key); // Normalize module names for comparison

      const remoteModuleName = remoteModuleIdToName.get(remoteModuleId) || remoteModuleId;
      const currentModuleName = moduleIdToName.get(currentModuleId) || currentModuleId; // Only consider a key moved if it's actually in a different module

      const moduleNameComparison = remoteModuleName !== currentModuleName;

      if (moduleNameComparison) {
        trulyMovedKeys.add(key);
        keyMovementDetails.set(key, {
          from: remoteModuleId,
          fromName: remoteModuleName,
          to: currentModuleId,
          toName: currentModuleName,
          type: 'module_change'
        });
      }
    } // Get keys from unmatched modules that exist in the current build


    const unmatchedRemoteModules = [...remoteModules.keys()].filter(id => !moduleEquivalenceMap.has(id));
    unmatchedRemoteModules.forEach(moduleId => {
      const module = remoteModules.get(moduleId);
      const moduleName = remoteModuleIdToName.get(moduleId) || moduleId;
      module.keys.forEach(key => {
        // Only add the key if it exists in the current build
        if (currentBuildKeys.has(key)) {
          const currentModuleId = currentKeyToModule.get(key);
          const currentModuleName = moduleIdToName.get(currentModuleId) || currentModuleId; // Only add if actually in a different module

          if (moduleName !== currentModuleName) {
            trulyMovedKeys.add(key);
            keyMovementDetails.set(key, {
              from: moduleId,
              fromName: moduleName,
              to: currentModuleId,
              toName: currentModuleName,
              type: 'unmatched_module'
            });
          }
        }
      });
    });
    const movedKeyCount = trulyMovedKeys.size;
    console.log(`Found ${extraKeys.size} new keys and ${movedKeyCount} moved keys`); // Add extra and moved keys to the plugin's tracking for master chunk inclusion

    plugin.extraKeys = extraKeys;
    plugin.movedKeys = trulyMovedKeys; // Store key counts for reference

    plugin.keyStats = {
      current: currentKeyCount,
      remote: remoteKeyCount,
      extra: extraKeys.size,
      moved: movedKeyCount
    }; // Update master keys in the generated meta manifest

    if (plugin.generatedMetaManifest) {
      // Only include extra (new) and moved keys in the master keys
      const masterKeys = [...Array.from(extraKeys), ...Array.from(trulyMovedKeys)];
      plugin.generatedMetaManifest.masterKeys = masterKeys;
    }

    return {
      extraKeys,
      movedKeys: trulyMovedKeys,
      keyStats: plugin.keyStats
    };
  } catch (error) {
    console.error(`Error during manifest comparison: ${error.message}`); // Return empty sets for keys if there was an error

    plugin.extraKeys = new Set();
    plugin.movedKeys = new Set();
    return {
      extraKeys: new Set(),
      movedKeys: new Set(),
      error: error.message
    };
  }
}